#!/bin/ksh

# Two sample korn shell scripts are provided which show how an embarrassingly parallel job 
# can be run as a LoadLeveler job. The first is autonomous_master.ksh which runs on the 
# master node and uses llspawn.stdio to start a second script, autonomous_slave.ksh on each 
# allocated node. 

# This shell script will use the LOADL_HOSTFILE environment variable to 
# locate the file containing machines allocated for the job. For each 
# machine in the file, the script will use the LoadLeveler llspawn.stdio 
# binary to start the autonomous_slave.ksh shell script on that node. 
  
# Either the specified command string and command count or the command file 
# will be passed as arguments to autonomous_slave.ksh. autonomous_master.ksh 
# will construct a string of task ids to pass as an argument to autonomous_slave.ksh.
# 
# The task id string is constructed for each machine by:
# 1. Sequentially assigning tasks ids to the  machine based on the number of tasks 
#    the machine can support.
# 2. Calculating the number of tasks a machine can support by counting the number of 
#    entries for the same machine in the file of machines.
  
# autonomous_master.ksh will also pass the LOADL_STEP_OUT and LOADL_STEP_ERR environment 
# variables as arguments to autonomous_slave.ksh. autonomous_slave.ksh uses these 
# are used by the autonomous_slave.ksh script to generate stdout and stderr filenames for 
# each task.

# autonomous_master.ksh options
#   -h Provides a short usage message.
#   -c Specifies a command string to invoke for each parallel task.
#   -n Specifies the number of tasks that will run the command string. 
#      If the number specified exceeds the number of tasks allocated for the job, 
#      the shell script will re-use task initiators as work on them completes.
#   -f Specifies a file containing commands (one per line) to invoke for each parallel task. 
#      If the number of commands in the file exceeds the number of tasks allocated for the job,  
#      the shell script will re-use task initiators as work on them completes.

# Parse command options 
while getopts c:n:f:hv o
do case "$o" in
    c)  COMMAND="$OPTARG";;
    n)  CMDCOUNT="$OPTARG";;
    f)  CMDFILE="$OPTARG";;
    h)  echo "Usage: $0 { -h | [-v] { {-c <command string> -n <command count>} | -f <command file name> } }" && exit 1;;
    v)  verbose="yes";;
    \?)  echo "Usage: $0 { -h | [-v] { {-c <command string> -n <command count>} | -f <command file name> } }" && exit 1;;
esac
done

if [[ "$verbose" == "yes" ]]; then echo "Entering $0";fi

integer cmd_cnt=0 # Keeps count of commands that need to be executed
integer mach_cnt=0 # Keeps count of unique machines in machine file
integer mach_lines=0 # Keeps total of machines in machine file (including duplicates)
integer i=0
integer machine_index=0
integer tid=0 # Keeps track of task ids when assigning tasks to machines
integer initiators_used=0 # Keeps track of initiators used for a machine

if [[ ! -z "$COMMAND" && "$verbose" == "yes" ]]
then
	echo "command option -c specified: ${COMMAND}\n"
fi
if [[ ! -z "$CMDCOUNT" && "$verbose" == "yes" ]]
then
	echo "command count option -n specified: ${CMDCOUNT}\n"
fi
if [[ ! -z "$CMDFILE" && "$verbose" == "yes" ]]
then
	echo "command file option -f specified: ${CMDFILE}\n"
fi

# Not allowed to specify both a command string and command file
if [[ ! -z "$COMMAND"  &&  ! -z "$CMDFILE" ]] 
then
	echo "Cannot specify both -f and -c"
	echo "Usage: $0 { -h | [-v] { {-c <command string> -n <command count>} | -f <command file name> } }" && exit 1
fi 

# Base for stderr files is provided through LoadLeveler runtime environment variable
if [[ "${LOADL_STEP_ERR:0:1}" = "/" ]] # If first character is not /, pre-pend pwd
then 
	STDERRBASE=$LOADL_STEP_ERR
else
	STDERRBASE=$PWD/$LOADL_STEP_ERR
fi
# Base for stdout files is provided through LoadLeveler runtime environment variable
if [[ "${LOADL_STEP_OUT:0:1}" = "/" ]] # If first character is not /, pre-pend pwd
then 
	STDOUTBASE=$LOADL_STEP_OUT
else
	STDOUTBASE=$PWD/$LOADL_STEP_OUT
fi
# Machine file is provided through LoadLeveler runtime environment variable
MACHFILE=$LOADL_HOSTFILE

if [[ -z "$MACHFILE" ]] 
then
	echo "LOADL_HOSTFILE environment variable is not set."
	echo "Usage: $0 { -h | [-v] { {-c <command string> -n <command count>} | -f <command file name> } }" && exit 1
fi

# Read machines into an array. Count machines and initiators per machine.
integer icounter=0

while read line # Read a line from the machine file
do
	if [[ $mach_cnt -eq 0  ||  "${mach[mach_cnt-1]}" != "${line}" ]]
	then
		# For each unique machine in the machine file
		if [ $mach_cnt -gt 0 ]
		then
			# Save the initiator count for the previous unique machine
			initiator_count[mach_cnt-1]=$icounter
		fi
		icounter=1 # Initialize the initiator count for the current machine
		mach[mach_cnt]=$line # Store the machine name in the mach array
		mach_cnt=mach_cnt+1 # Count the unique machine
	else
		# For each duplicate machine in the machine file
		icounter=icounter+1 # Increment the count of initiators for this machine
	fi
	mach_lines=mach_lines+1 # Count number of lines in machine file
done < "$MACHFILE"
initiator_count[mach_cnt-1]=$icounter # Save the initiator count for the last unique machine

if [[ -z "$CMDCOUNT" ]] 
then
	# If command count was not specified, use the number of lines in the machine file
	CMDCOUNT=mach_lines;
fi

i=0
if [[ "$verbose" == "yes" ]]
then
	#
	# Display all machines that were just read
	#
	while [ "$i" -lt "$mach_cnt" ]
	do
	        echo "Machine: ${mach[$i]} initiators: ${initiator_count[$i]}"
	        i=i+1
	done
        echo "\n"
fi

if [[ ! -z "$COMMAND" ]] 
then
	cmd_cnt=$CMDCOUNT

	if [[ "$verbose" == "yes" ]]; then echo "Entering $0 shell script, COMMAND=$COMMAND MACHFILE=$MACHFILE STDOUTBASE=$STDOUTBASE STDERRBASE=$STDERRBASE"; fi
else
	cmd_cnt=`cat $CMDFILE | wc -l`
	if [[ "$verbose" == "yes" ]]; then echo "Entering $0 shell script, CMDFILE=$CMDFILE MACHFILE=$MACHFILE STDOUTBASE=$STDOUTBASE STDERRBASE=$STDERRBASE"; fi
fi

# Invoke the autonomous_slave.ksh script on each machine in the machine list
# Continue until all commands have run. 

machine_index=0
tid=0
	
# Assign tasks to machines to handle the count of commands
while [ "$tid" -lt "$cmd_cnt" ]
do
	# Construct a string of task ids for the machine
	initiators_used=0
	icounter=${initiator_count[$machine_index]} # Get the initiator count for the machine

	# Assign as many tasks to this machines as there are initiators
	while [ "$initiators_used" -lt "$icounter" ]
	do
		if [ "$initiators_used" -eq 0 ]
		then
			tid_list=$tid
		else
			tid_list="$tid_list $tid"
		fi 
		initiators_used=initiators_used+1 # increment intiatiors used for this machine
		tid=tid+1 # increment overall task id
		if [ "$tid" -ge "$cmd_cnt" ] # Need to break out if we handled enough commands
		then
			break
		fi
	done


	# Form the command line to invoke autonomous_slave.ksh on a machine in the machine list
	if [[ "$verbose" == "yes" ]]; then dashv="-v";fi # Pass -v flag to autonomous_slave.ksh
	if [[ ! -z "$COMMAND" ]] 
	then
		COMMANDLINE="/opt/ibmll/LoadL/resmgr/full/samples/autonomous/autonomous_slave.ksh -t \"${tid_list}\" -o \"${STDOUTBASE}\" -e \"${STDERRBASE}\" -c \"${COMMAND}\" $dashv"
	else
		COMMANDLINE="/opt/ibmll/LoadL/resmgr/full/samples/autonomous/autonomous_slave.ksh -t \"${tid_list}\" -o \"${STDOUTBASE}\" -e \"${STDERRBASE}\" -f \"${CMDFILE}\" $dashv"
	fi

	# Use llspawn.stdio to invoked the command line on the remote machine
	if [[ "$verbose" == "yes" ]]; then echo $COMMANDLINE; fi
	llspawn.stdio ${mach[$machine_index]} "${COMMANDLINE}" &

	machine_index=machine_index+1 # increment to the next machine to use

	# Once all machines are used, if there are commands left to run,
	# wait until tasks finish before continuing
	if [[ "$machine_index" -eq "$mach_cnt"  &&  "$tid" -lt "$cmd_cnt" ]]
	then
		if [[ "$verbose" == "yes" ]]; then echo "Waiting for child processes (within loop)"; fi
		wait
		machine_index=0
	fi
done

# If there are any machines in use, wait until tasks finish before exiting
if [ "$machine_index" -ne 0 ]
then
	if [[ "$verbose" == "yes" ]]; then echo "Waiting for child processes (after loop)"; fi
	wait
	machine_index=0
fi

if [[ "$verbose" == "yes" ]]; then echo "Exiting $0";fi

