#!/bin/ksh 

# Two sample korn shell scripts are provided which show how an embarrassingly parallel job 
# can be run as a LoadLeveler job. The second is autonomous_slave.ksh which is started on 
# each allocated node by the llspawn.stdio command invoked by autonomous_master.ksh.
# allocated node. autonomouss_slave.ksh starts the tasks designated to run on a node. 

# This shell script will invoke commands specified by the input arguments.
# - When the -c option is used, the command string specified is invoked for each task 
#   id in the task id string, specified with option  t.
# - When the -f option is used, the task ids, in the task id string specified with option t, 
#   are used to select command lines from the file to invoke.

# Prior to invoking each command the task id is exported in the LL_SAMPLE_TASK_ID environment 
# variable. Having access to the task id allows each task to determine what work it has to do.

# Standard out and standard error from each command is written to separate files using 
# filenames constructed from the -o and -e option values and the LL_SAMPLE_TASK_ID environment 
# variable.

# autonomous_slave.ksh options:
# -h Provides a short usage message.
# -c Specifies a command string to invoke for each parallel task.
# -f Specifies a file containing commands (one per line) to invoke for each parallel task. The 
#    task ids, specified in the task id string with option t, identify which command lines from 
#    the file to invoke on the node.
# -t Specifies a string containing a blank delimited list of task id numbers to use when starting 
#    tasks on the node.
# -o Specifies the stdout base filename is used to construct each task specific stdout filename.
# -e Specifies the stderr base filename is used to construct each task specific stderr filename.

# Parse command options 
while getopts c:f:m:o:e:t:hv o
do case "$o" in
    c)  COMMAND="$OPTARG";;
    f)  CMDFILE="$OPTARG";;
    o)  STDOUTBASE="$OPTARG";;
    e)  STDERRBASE="$OPTARG";;
    t)  TASKIDS="$OPTARG";;
    h)  echo "Usage: $0 { -h | [-v] {-c command string | -f filename} { -t task id string} {-o <stdout base filename>} {-e <stderr base filename>} }" && exit 1;;
    v)  verbose="yes";;
    \?)  echo "Usage: $0 { -h | [-v] {-c command string | -f filename} { -t task id string} {-o <stdout base filename>} {-e <stderr base filename>} }" && exit 1;;
esac
done

if [[ "$verbose" == "yes" ]]; then echo "Entering $0 on `hostname -s`";fi

integer cmd_cnt=0
integer mach_cnt=0
integer i=0
integer j=0
integer k=0
integer last_index=1
integer task_index=1


if [[ ! -z "$COMMAND" && "$verbose" == "yes" ]]
then
	echo "command: $COMMAND"
fi
if [[ ! -z "$CMDFILE" && "$verbose" == "yes" ]]
then
	echo "CMDFILE: $CMDFILE"
fi
if [[ -z "$TASKIDS" ]] 
then
	echo "task id string not specified"
	echo "Usage: $0 { -h | [-v] {-c command string | -f filename} { -t task id string} {-o <stdout base filename>} {-e <stderr base filename>} }" && exit 1
fi
if [[ -z "$STDERRBASE" ]] 
then
	echo "stderr base file name not specified"
	echo "Usage: $0 { -h | [-v] {-c command string | -f filename} { -t task id string} {-o <stdout base filename>} {-e <stderr base filename>} }" && exit 1
fi
if [[  -z "$STDOUTBASE" ]] 
then
	echo "stdout base file name not specified"
	echo "Usage: $0 { -h | [-v] {-c command string | -f filename} { -t task id string} {-o <stdout base filename>} {-e <stderr base filename>} }" && exit 1
fi

# Not allowed to specify both a command string and command file
if [[ ! -z "$COMMAND"  &&  ! -z "$CMDFILE" ]] 
then
	echo "Cannot specify both -f and -c"
	echo "Usage: $0 { -h | [-v] {-c command string | -f filename} { -t task id string} {-o <stdout base filename>} {-e <stderr base filename>} }" && exit 1
fi 

if [[ ! -z "$COMMAND" ]] 
then
	# If a command string was specified, invoke it for each task in the task id string 
	if [[ "$verbose" == "yes" ]]; then echo "COMMAND=$COMMAND STDOUTBASE=$STDOUTBASE STDERRBASE=$STDERRBASE";fi

	last_index=`echo $TASKIDS | awk '{print NF}'` # Get the index to the last task id from the string of task ids
	task_index=1
	while [[ $task_index -le $last_index ]] # Need to start a command for each task id in the string of task ids
	do
		tid=`echo $TASKIDS | cut -f${task_index} -d' '` # Get a task id
		if [ ! -z $tid ] 
		then 
			if [[ "$verbose" == "yes" ]]; then echo "tid = $tid";fi
			export LL_SAMPLE_TASK_ID=$tid # export and environment variable containing the task id

			# Invoke the passed command. Re-direct stdout and sterr to files with the task id as the suffix.
			if [[ "$verbose" == "yes" ]]; then echo "${COMMAND} > ${STDOUTBASE}.$tid  2> ${STDERRBASE}.$tid";fi
			ksh -c "${COMMAND}" > ${STDOUTBASE}.$tid  2> ${STDERRBASE}.$tid &
		else
			if [[ "$verbose" == "yes" ]]; then echo "tid is empty";fi
			break
		fi
		task_index=task_index+1 # Advance to the next task id
	done

	# Wait for all tasks to terminate
	if [[ "$verbose" == "yes" ]]; then echo "Waiting for child processes (after loop)";fi
	wait
else
	# If a command file was specified, invoke selected commands from the file on based on the task id string
	if [[ "$verbose" == "yes" ]]; then echo "CMDFILE=$CMDFILE STDOUTBASE=$STDOUTBASE";fi

	# Read commands into an array
	cmd_cnt=0;
	while read line
	do
		cmd[cmd_cnt]=$line
		cmd_cnt=cmd_cnt+1
	done < "$CMDFILE"

	# Use each task id as an index into the command array to choose the command to invoke
	last_index=`echo $TASKIDS | awk '{print NF}'` # Get the index to the last task id from the string of task ids
	task_index=1
	while [[ $task_index -le $last_index ]] # Need to start a command for each task id in the string of task ids
	do
		tid=`echo $TASKIDS | cut -f${task_index} -d' '` # Get a task id
		if [ ! -z $tid ] 
		then 
			if [[ "$verbose" == "yes" ]]; then echo "tid = $tid";fi
			export LL_SAMPLE_TASK_ID=$tid # export and environment variable containing the task id

			# Retrieve the command corresonding to the task id from the array of commands.
			# Invoke the retrieved command. Re-direct stdout and sterr to files with the task id as the suffix.
			if [[ "$verbose" == "yes" ]]; then echo "${cmd[$tid]} > ${STDOUTBASE}.$tid  2> ${STDERRBASE}.$tid";fi
			ksh -c "${cmd[$tid]}" > ${STDOUTBASE}.$tid  2> ${STDERRBASE}.$tid &
		else
			break
		fi
		task_index=task_index+1 # Advance to the next task id
	done

	# Wait for all tasks to terminate
	if [[ "$verbose" == "yes" ]]; then echo "Waiting for child processes (after loop)";fi
	wait
fi

if [[ "$verbose" == "yes" ]]; then echo "Exiting $0 on `hostname -s`";fi

